from datetime import datetime, timezone
from unittest.mock import patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.ponos.models import State, Task
from arkindex.process.models import Process, ProcessMode
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import User


class TestProcessList(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        user2 = User.objects.create_user("user2@test.test", display_name="Process creator")
        cls.files_process = cls.corpus.processes.create(
            mode=ProcessMode.Files,
            creator=cls.user,
        )
        cls.workers_process = cls.corpus.processes.get(mode=ProcessMode.Workers)
        cls.workers_process_user2 = cls.corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=user2,
        )

        cls.other_corpus = Corpus.objects.create(name="Other corpus")
        cls.other_process = cls.other_corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=user2,
        )

        cls.export_process = cls.corpus.processes.create(
            mode=ProcessMode.Export,
            creator=cls.user
        )

    def test_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:process-list"))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list(self):
        """
        A user can list processes they have access to
        """
        # Order processes by causing updates on them
        with patch("arkindex.process.tasks.initialize_activity.delay"):
            self.other_process.run()
        self.files_process.save()
        self.workers_process_user2.save()
        # An update on a process task will make it first
        task = self.other_process.tasks.first()
        task.save()
        self.assertTrue(self.user.processes.filter(mode=ProcessMode.Local).exists())
        self.client.force_login(self.user)

        workers_process_user2_response = {
            "id": str(self.workers_process_user2.id),
            "name": None,
            "mode": "workers",
            "chunks": 1,
            "corpus": str(self.corpus.id),
            "state": "unscheduled",
            "activity_state": "disabled",
            "created": self.workers_process_user2.created.isoformat().replace("+00:00", "Z"),
            "creator": "Process creator",
            "updated": self.workers_process_user2.updated.isoformat().replace("+00:00", "Z"),
            "started": None,
            "finished": None,
            "use_cache": False,
        }
        files_process_response = {
            "id": str(self.files_process.id),
            "name": None,
            "mode": "files",
            "chunks": 1,
            "corpus": str(self.corpus.id),
            "state": "unscheduled",
            "activity_state": "disabled",
            "created": self.files_process.created.isoformat().replace("+00:00", "Z"),
            "creator": "Test user",
            "updated": self.files_process.updated.isoformat().replace("+00:00", "Z"),
            "started": None,
            "finished": None,
            "use_cache": False,
        }
        other_process_response = {
            "id": str(self.other_process.id),
            "name": None,
            "mode": "workers",
            "chunks": 1,
            "corpus": str(self.other_corpus.id),
            "state": "unscheduled",
            "activity_state": "pending",
            "created": self.other_process.created.isoformat().replace("+00:00", "Z"),
            "creator": "Process creator",
            "updated": self.other_process.updated.isoformat().replace("+00:00", "Z"),
            "started": self.other_process.started.isoformat().replace("+00:00", "Z"),
            "finished": None,
            "use_cache": False,
        }
        workers_process_response = {
            "id": str(self.workers_process.id),
            "name": "Process fixture",
            "mode": "workers",
            "chunks": 1,
            "corpus": str(self.corpus.id),
            "state": "unscheduled",
            "activity_state": "disabled",
            "created": self.workers_process.created.isoformat().replace("+00:00", "Z"),
            "creator": "Test user",
            "updated": self.workers_process.updated.isoformat().replace("+00:00", "Z"),
            "started": None,
            "finished": None,
            "use_cache": False,
        }
        export_process_response = {
            "id": str(self.export_process.id),
            "name": None,
            "mode": "export",
            "chunks": 1,
            "corpus": str(self.corpus.id),
            "state": "unscheduled",
            "activity_state": "disabled",
            "created": self.export_process.created.isoformat().replace("+00:00", "Z"),
            "creator": "Test user",
            "updated": self.export_process.updated.isoformat().replace("+00:00", "Z"),
            "started": None,
            "finished": None,
            "use_cache": False,
        }

        cases = [
            ({}, 5, [
                workers_process_user2_response,
                files_process_response,
                other_process_response,
                export_process_response,
                workers_process_response,
            ]),
            ({"mode": "files"}, 5, [
                files_process_response,
            ]),
            ({"mode": "workers"}, 5, [
                workers_process_user2_response,
                other_process_response,
                workers_process_response,
            ]),
            # There is a local process for the current user, but it is not visible through this endpoint
            # Two less queries because we only count processes and find that there are 0, so we don't retrieve them or their tasks
            ({"mode": "local"}, 3, []),
            ({"started": "true"}, 5, [
                other_process_response,
            ]),
            # One less query because non-started processes never have any tasks
            ({"started": "false"}, 4, [
                workers_process_user2_response,
                files_process_response,
                export_process_response,
                workers_process_response,
            ]),
            ({"created": "true"}, 5, [
                files_process_response,
                export_process_response,
                workers_process_response,
            ]),
            ({"created": "false"}, 5, [
                workers_process_user2_response,
                other_process_response,
            ]),
            ({"name": "proc"}, 5, [
                workers_process_response,
            ]),
            ({"corpus": str(self.corpus.id)}, 5, [
                workers_process_user2_response,
                files_process_response,
                export_process_response,
                workers_process_response,
            ]),
            ({"corpus": str(self.other_corpus.id)}, 5, [
                other_process_response,
            ]),
            ({"id": str(self.files_process.id)[:10]}, 5, [
                files_process_response,
            ]),
            ({"mode": "export"}, 5, [export_process_response])
        ]

        for filters, queries, results in cases:
            with self.subTest(**filters), self.assertNumQueries(queries):
                response = self.client.get(reverse("api:process-list"), filters)
                self.assertEqual(response.status_code, status.HTTP_200_OK)

                self.assertDictEqual(response.json(), {
                    "count": len(results),
                    "number": 1,
                    "next": None,
                    "previous": None,
                    "results": results,
                })

    def test_invalid_filters(self):
        self.client.force_login(self.user)

        cases = [
            ({"corpus": "oh-no"}, {"corpus": ["'oh-no' is not a valid UUID"]}),
            ({"mode": "unexisting_mode"}, {"mode": ["Mode 'unexisting_mode' does not exist"]}),
            ({"state": "spain"}, {"state": ["State 'spain' does not exist"]}),
        ]
        for filters, expected_errors in cases:
            with self.subTest(**filters), self.assertNumQueries(2):
                response = self.client.get(reverse("api:process-list"), filters)
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), expected_errors)

    def test_state_filter(self):
        """
        Only last run should be used to filter processes by state
        """
        completed_process = Process.objects.create(mode=ProcessMode.Workers, creator=self.user, corpus=self.corpus)
        completed_process.tasks.create(depth=0, run=0, slug="task", state=State.Failed, ttl=0)
        completed_process.tasks.create(depth=0, run=1, slug="task", state=State.Completed, ttl=0)
        self.assertEqual(completed_process.state, State.Completed)

        error_process = Process.objects.create(mode=ProcessMode.Workers, creator=self.user, corpus=self.corpus)
        error_process.tasks.create(depth=0, run=0, slug="error_task", state=State.Error, ttl=0)
        error_process.tasks.create(depth=0, run=0, slug="completed_task", state=State.Completed, ttl=0)
        self.assertEqual(error_process.state, State.Error)

        stopped_process = Process.objects.create(mode=ProcessMode.Workers, creator=self.user, corpus=self.corpus)
        stopped_process.tasks.create(depth=0, run=0, slug="completed_task", state=State.Completed, ttl=0)
        stopped_process.tasks.create(depth=0, run=1, slug="completed_task", state=State.Completed, ttl=0)
        stopped_process.tasks.create(depth=0, run=2, slug="stopped_task", state=State.Stopped, ttl=0)
        self.assertEqual(stopped_process.state, State.Stopped)

        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:process-list"), {"state": "completed"})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertCountEqual([r["id"] for r in response.json()["results"]], [str(completed_process.id)])

    def test_state_filter_stopped(self):
        stopping_process = Process.objects.create(mode=ProcessMode.Workers, creator=self.user, corpus=self.corpus)
        stopping_process.tasks.create(depth=0, run=0, slug="task", state=State.Error, ttl=0)
        stopping_process.tasks.create(depth=0, run=1, slug="task", state=State.Stopping, ttl=0)
        self.assertEqual(stopping_process.state, State.Stopping)

        error_process = Process.objects.create(mode=ProcessMode.Workers, creator=self.user, corpus=self.corpus)
        error_process.tasks.create(depth=0, run=0, slug="error_task", state=State.Error, ttl=0)
        error_process.tasks.create(depth=0, run=0, slug="stopping_task", state=State.Stopping, ttl=0)
        self.assertEqual(error_process.state, State.Error)

        stopped_process_1 = Process.objects.create(mode=ProcessMode.Workers, creator=self.user, corpus=self.corpus)
        stopped_process_1.tasks.create(depth=0, run=0, slug="unscheduled_task", state=State.Unscheduled, ttl=0)
        stopped_process_1.tasks.create(depth=0, run=0, slug="stopped_task", state=State.Stopped, ttl=0)
        self.assertEqual(stopped_process_1.state, State.Stopped)

        stopped_process_2 = Process.objects.create(mode=ProcessMode.Workers, creator=self.user, corpus=self.corpus)
        stopped_process_2.tasks.create(depth=0, run=0, slug="failed_task", state=State.Failed, ttl=0)
        stopped_process_2.tasks.create(depth=0, run=1, slug="pending_task", state=State.Pending, ttl=0)
        stopped_process_2.tasks.create(depth=0, run=1, slug="stopped_task", state=State.Stopped, ttl=0)
        self.assertEqual(stopped_process_2.state, State.Stopped)

        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:process-list"), {"state": "stopped"})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertCountEqual(
            [r["id"] for r in response.json()["results"]],
            [str(stopped_process_1.id), str(stopped_process_2.id)]
        )

    def test_state_filter_unscheduled(self):
        """
        Processes with no task should be interpreted as unscheduled
        """
        unscheduled_with_tasks = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=self.user,
            corpus=self.corpus,
        )
        unscheduled_with_tasks.tasks.create(depth=0, run=0, slug="task", state=State.Completed, ttl=0)
        unscheduled_with_tasks.tasks.create(depth=0, run=1, slug="task", state=State.Unscheduled, ttl=0)
        unscheduled_with_tasks.started = datetime.now(timezone.utc)
        unscheduled_with_tasks.save()
        self.assertEqual(unscheduled_with_tasks.state, State.Unscheduled)

        self.assertFalse(self.files_process.tasks.exists())
        self.assertEqual(self.files_process.state, State.Unscheduled)
        self.assertFalse(self.workers_process_user2.tasks.exists())
        self.assertEqual(self.workers_process_user2.state, State.Unscheduled)
        self.assertFalse(self.other_process.tasks.exists())
        self.assertEqual(self.other_process.state, State.Unscheduled)

        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:process-list"), {"state": "unscheduled", "started": True})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 1,
            "number": 1,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": str(unscheduled_with_tasks.id),
                    "name": None,
                    "mode": "workers",
                    "chunks": 1,
                    "corpus": str(self.corpus.id),
                    "state": "unscheduled",
                    "activity_state": "disabled",
                    "created": unscheduled_with_tasks.created.isoformat().replace("+00:00", "Z"),
                    "creator": "Test user",
                    "updated": unscheduled_with_tasks.updated.isoformat().replace("+00:00", "Z"),
                    "started": unscheduled_with_tasks.started.isoformat().replace("+00:00", "Z"),
                    "finished": None,
                    "use_cache": False
                },
            ]
        })

        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:process-list"), {"state": "unscheduled", "started": False})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "count": 5,
            "number": 1,
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": str(self.export_process.id),
                    "name": None,
                    "mode": "export",
                    "chunks": 1,
                    "corpus": str(self.corpus.id),
                    "state": "unscheduled",
                    "activity_state": "disabled",
                    "created": self.export_process.created.isoformat().replace("+00:00", "Z"),
                    "creator": "Test user",
                    "updated": self.export_process.updated.isoformat().replace("+00:00", "Z"),
                    "started": None,
                    "finished": None,
                    "use_cache": False,
                },
                {
                    "id": str(self.other_process.id),
                    "name": None,
                    "mode": "workers",
                    "chunks": 1,
                    "corpus": str(self.other_corpus.id),
                    "state": "unscheduled",
                    "activity_state": "disabled",
                    "created": self.other_process.created.isoformat().replace("+00:00", "Z"),
                    "creator": "Process creator",
                    "updated": self.other_process.updated.isoformat().replace("+00:00", "Z"),
                    "started": None,
                    "finished": None,
                    "use_cache": False
                },
                {
                    "id": str(self.workers_process_user2.id),
                    "name": None,
                    "mode": "workers",
                    "chunks": 1,
                    "corpus": str(self.corpus.id),
                    "state": "unscheduled",
                    "activity_state": "disabled",
                    "created": self.workers_process_user2.created.isoformat().replace("+00:00", "Z"),
                    "creator": "Process creator",
                    "updated": self.workers_process_user2.updated.isoformat().replace("+00:00", "Z"),
                    "started": None,
                    "finished": None,
                    "use_cache": False
                },
                {
                    "id": str(self.files_process.id),
                    "name": None,
                    "mode": "files",
                    "chunks": 1,
                    "corpus": str(self.corpus.id),
                    "state": "unscheduled",
                    "activity_state": "disabled",
                    "created": self.files_process.created.isoformat().replace("+00:00", "Z"),
                    "creator": "Test user",
                    "updated": self.files_process.updated.isoformat().replace("+00:00", "Z"),
                    "started": None,
                    "finished": None,
                    "use_cache": False
                },
                {
                    "id": str(self.workers_process.id),
                    "name": "Process fixture",
                    "mode": "workers",
                    "chunks": 1,
                    "corpus": str(self.corpus.id),
                    "state": "unscheduled",
                    "activity_state": "disabled",
                    "created": self.workers_process.created.isoformat().replace("+00:00", "Z"),
                    "creator": "Test user",
                    "updated": self.workers_process.updated.isoformat().replace("+00:00", "Z"),
                    "started": None,
                    "finished": None,
                    "use_cache": False
                },
            ]
        })

    def test_template_mode_ignores_with_tasks(self):
        """
        The 'with_tasks' filter should be ignored and some templates should be returned
        when filtering on the template process mode, because templates cannot have tasks
        """
        self.corpus.processes.create(mode=ProcessMode.Template, creator=self.user)
        self.assertFalse(Task.objects.filter(process__mode=ProcessMode.Template).exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse("api:process-list"),
                data={"mode": "template", "with_tasks": True},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data["count"], 1)
        self.assertNotEqual(data["results"], [])
